
/* 
 * Test of the MI CCD driver.
 * Copyright (C) 2010-2015 Petr Kubanek <petr@kubanek.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <errno.h>
#include <getopt.h>
#include <malloc.h>
#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>

#ifdef HAVE_FITSIO
#include <fitsio.h>
#include <longnam.h>
#endif

#include "miccd.h"

camera_info_t camera_info;
camera_t camera;

int img_c = 1;

// functions for image processing

#ifdef HAVE_FITSIO
void print_fits_status (const char *msg, const char *fn, int fits_status)
{
	char fe[200];
	char fm[81];

	fits_get_errstatus (fits_status, fe);
	fits_read_errmsg (fm);

	fprintf (stderr, "%s %s: %s message: %s\n", msg, fn, fe, fm);
}
#endif

void process_buf (uint16_t *buf, int write_fits)
{
	double sum = 0;
	int j;

	for (j = 0; j < (camera_info.w * camera_info.h); j++)
	{
		sum += (buf)[j];
	}
	printf ("sum %f average %f\n", sum, ((double) sum) / (camera_info.w * camera_info.h));

	if (write_fits)
	{
		char fn[17];

		snprintf (fn, 16, "!mi_%06d.fits", img_c++);

#ifdef HAVE_FITSIO
		int fits_status;
		fitsfile *ffile = NULL;

		long sizes[2];

		sizes[0] = camera_info.w;
		sizes[1] = camera_info.h;

		fits_create_file (&ffile, fn, &fits_status);
		fits_create_img (ffile, USHORT_IMG, 2, sizes, &fits_status);

		fits_update_key (ffile, TINT, "IMGNUM", (void *) &img_c, "image number in MI CCD test sequence", &fits_status);
		fits_write_history (ffile, "Created with MI test program (with CFITSIO)", &fits_status);

		fits_write_img (ffile, TUSHORT, 1, sizes[0] * sizes[1], buf, &fits_status);

		fits_close_file (ffile, &fits_status);

		if (fits_status)
		{
			print_fits_status ("cannot write data to", fn + 1, fits_status);
			return;
		}
#else
		FILE *f = fopen (fn + 1, "w");
		if (f == NULL)
		{
			perror ("cannot open");
			return;
		}

		// write FITS header
		fprintf (f,	"SIMPLE  =                    T / file does conform to FITS standard             "
				"BITPIX  =                   16 / number of bits per data pixel                  "
				"NAXIS   =                    2 / number of data axes                            " 
				"NAXIS1  =               %6d / length of data axis 1                          "
				"NAXIS2  =               %6d / length of data axis 1                          "
				"EXTEND  =                    T / FITS dataset may contain extensions            "
				"COMMENT   FITS (Flexible Image Transport System) format is defined in 'Astronomy"
				"COMMENT   and Astrophysics', volume 376, page 359; bibcode: 2001A&A...376..359H "
				"BZERO   =                32768 / offset data range to that of unsigned short    "
				"BSCALE  =                    1 / default scaling factor                         "
				"IMGNUM  =                %5d / image number in MI CCD test sequence           "
				"HISTORY Created with MI test program (without CFITSIO)                          "
				"END                                                                             "
			, camera_info.w, camera_info.h, img_c);

		for (j = 13; j < 36; j++)
			fprintf (f, "                                                                                ");

		fwrite (buf, camera_info.w * camera_info.h, 2, f);
		fclose (f);
#endif
		printf ("Image data writen to %s\n", fn + 1);
	}
}

// example of exposure code

int expose_g1 (float sec, int bit16, int lownoise, char *buf)
{
	int ret;
	ret = miccd_clear (&camera);
	if (ret)
		return ret;

	ret = miccd_g1_mode (&camera, bit16, lownoise);
	if (ret)
		return ret;

	if (sec <= 8)
	{
		ret = miccd_start_exposure (&camera, 0, 0, camera_info.w, camera_info.h, sec);
		if (ret < 0)
			return ret;

		usleep (sec * 1000000);
	}
	else
	{
		usleep (sec * 1000000);

		ret = miccd_start_exposure (&camera, 0, 0, camera_info.w, camera_info.h, -1);
		if (ret)
			return ret;
	}

	// get data..
	ret = miccd_read_data (&camera, 2 * camera_info.w * camera_info.h, (char *) buf, camera_info.w, camera_info.h);
	return ret;
}


int expose_g2 (float sec, int dark, uint8_t mode, char *buf)
{
	int ret;
	ret = miccd_clear (&camera);
	if (ret)
		return ret;

	ret = miccd_hclear (&camera);
	if (ret)
		return ret;

	if (dark == 0)
	{
		printf ("opening shutter\n");

		ret = miccd_open_shutter (&camera);
		if (ret)
			return ret;
	}

	usleep (sec * 1000000);

	if (dark == 0)
	{
		printf ("closing shutter\n");

		ret = miccd_close_shutter (&camera);
		if (ret)
			return ret;
	}

	ret = miccd_shift_to0 (&camera);
	if (ret)
		return ret;

	ret = miccd_mode (&camera, mode);
	if (ret)
		return ret;

	// get data..
	ret = miccd_read_frame (&camera, 1, 1, 0, 0, camera_info.w, camera_info.h, (char *) buf);

	return ret;
}

int main (int argc, char **argv)
{
	float valf;
	int ret;

	int8_t mode = 0;
	int product_id = 0;

	int dark_count = 1;
	int light_count = 1;

	int bit16 = 0;
	int lownoise = 0;

	float exposure = 1;

	int write_fits = 0;

	// parse options
	while (1)
	{
		int option_index = 0;
		static struct option long_options[] = {
			{"mode",        required_argument, 0, 'm'},
			{"bit16",       no_argument,       0, 'b'},
			{"lownoise",    no_argument,       0, 'n'},
			{"exposure",    required_argument, 0, 'e'},
			{"light-count", required_argument, 0, 'l'},
			{"dark-count",  required_argument, 0, 'd'},
			{"write-fits",  no_argument,       0, 'w'},
			{"help",        no_argument,       0, 'h'},
			{0,             0,                 0,  0 }
		};

		char c = getopt_long(argc, argv, "m:bne:d:l:wh", long_options, &option_index);
		if (c == -1)
			break;

		switch (c)
		{
			case 'm':
				mode = atoi(optarg);
				if (mode < 0 || mode > 2)
				{
					fprintf (stderr, "mode argument must be between 0 and 2");
					exit (1);
				}
				break;
			case 'b':
				bit16 = 1;
				break;
			case 'n':
				lownoise = 1;
				break;
			case 'e':
				exposure = atof (optarg);
				if (exposure <= 0)
				{
					fprintf (stderr, "Exposure argument must be a positive float number");
					exit (1);
				}
				break;
			case 'd':
				dark_count = atoi (optarg);
				break;
			case 'l':
				light_count = atoi (optarg);
				break;
			case 'w':
				write_fits = 1;
				break;
			case 'h':
				fprintf (stdout, "Simple demonstration program for Moravian Instruments CCD\n\n"
					"Options:\n"
					"\t--mode,-m         Camera mode. 0 = NORMAL, 1 = LOW NOISE, 2 = ULTRA LOW NOISE. 0 by default. Available only for G2+ cameras.\n"
					"\t--bit16,-b        16 bit mode. Available only for G0 and G1 models.\n"
					"\t--lownoise,-n     Low noise mode. Available only for G0 and G1 models.\n"
					"\t--exposure,-e     Exposure time in seconds (and fractions of seconds). Default to 1.\n"
					"\t--dark-count,-d   Number of dark frames produced. Default to 1, available only for G2+ cameras. Use negative number to obtain arbitary number of dark frames.\n"
					"\t--light-count,l   Number of light frames produced. Default to 1. Use negative number to obtain arbitary number of light frames.\n"
					"Arguments:\n"
					"\tProduct ID. If not specified, 0 is assument, which will select the first camera found. Otherwise\n"
					"Example:\n"
					"\t# produce 10 dark and 10 light images in low noise mode\n"
					"\t%1$s -d 10 -l 10 -m 1\n\n"
					"\t# produce arbitary number of 0.2 seconds light images images in ultra low noise mode with camera with product ID 2432\n"
					"\t%1$s -l -1 -m 2 -e 0.2 2432\n\n",
					argv[0]);
				exit (0);
				break;
		}
	}

	if (optind > (argc + 1))
	{
		fprintf (stderr, "Product ID can be specified as only argument!\n");
		return -1;
	}
	else if (optind + 1 == argc)
	{
		product_id = atoi (argv[optind]);
	}

	ret = miccd_open (product_id, &camera);

	if (ret < 0)
	{
		fprintf (stderr, "cannot open camera with product id %s : %s\n", argv[1], strerror (errno));
		return -1;
	}

	ret = miccd_info (&camera, &camera_info);
	if (ret)
	{
		fprintf (stderr, "cannot call &camera info %d\n", ret);
		return -1;
	}

	printf ("Camera product ID.....%d\n", camera_info.id);
	printf ("HW revision...........%d\n", camera_info.hwrevision);
	printf ("Description...........%14s\n", camera_info.description);
	printf ("Serial................%14s\n", camera_info.serial);
	printf ("Chip..................%14s\n", camera_info.chip);
	printf ("Size..................%d x %d\n", camera_info.w, camera_info.h);
	printf ("Pixel size............%d x %d\n", camera_info.pw, camera_info.ph);

	switch (camera.model)
	{
		case G10300:
		case G10400:
		case G10800:
		case G11200:
		case G11400:
		case G12000:
			printf ("16 bit................%s\n", bit16 ? "yes" : "no");
			printf ("Low noise.............%s\n", lownoise ? "yes" : "no");
			ret = miccd_g1_mode (&camera, bit16, lownoise);
			if (ret)
			{
				fprintf (stderr, "cannot set MI G1 mode to 0, exiting\n");
				return ret;
			}
			break;
		case G2:
		case G3:
		case G3_H:
			{
				static char* mode_string[] = {"NORMAL", "LOW NOISE", "ULTRA LOW NOISE"};
				printf ("Camera mode...........%s\n", mode_string[mode]);
				ret = miccd_mode (&camera, 0);
				if (ret)
				{
					fprintf (stderr, "cannot set MI mode to 0, exiting\n");
					return ret;
				}
			}
	}

	ret = miccd_chip_temperature (&camera, &valf);
	if (ret)
		return ret;
	
	printf ("CCD temperature.......%f\n", valf);

	switch (camera.model)
	{
		case G10300:
		case G10400:
		case G10800:
		case G11200:
		case G11400:
		case G12000:
			break;
		case G2:
		case G3:
		case G3_H:
			ret = miccd_environment_temperature (&camera, &valf);
			if (ret)
				return ret;
			printf ("Outside temperature...%f\n", valf);
	}

	char *buf = (char *) malloc (camera_info.w * camera_info.h * 2);

	switch (camera.model)
	{
		case G10300:
		case G10400:
		case G10800:
		case G11200:
		case G11400:
		case G12000:
			for (; light_count != 0; light_count > 0 ? light_count-- : light_count)
			{
				ret = expose_g1 (exposure, bit16, lownoise, buf);
				if (ret)
				{
					fprintf (stderr, "cannot expose light frame\n");
					return ret;
				}
				process_buf ((uint16_t *) buf, write_fits);
			}
			break;
		case G2:
		case G3:
		case G3_H:
			for (; light_count != 0; light_count > 0 ? light_count-- : light_count)
			{
				ret = expose_g2 (exposure, 0, mode, buf);
				if (ret)
				{
					fprintf (stderr, "cannot expose light frame\n");
					return ret;
				}
				process_buf ((uint16_t *) buf, write_fits);
			}
			for (; dark_count != 0; dark_count > 0 ? dark_count-- : dark_count)
			{
				ret = expose_g2 (exposure, 1, mode, buf);
				if (ret)
				{
					fprintf (stderr, "cannot expose dark frame\n");
					return ret;
				}
				process_buf ((uint16_t *) buf, write_fits);
			}
			break;
	}

	free (buf);

	return 0;
}
